//
// Visualization and Cluster Analysis of well-log data stored in plain text files. 
// Allows to group together data points with similar log responses by applying
// K-means clustering technique and makes graphical display of clustering results.
// The user is provided with a choice of different clustering algorithms, number
// of clusters and colors to be assigned with the clusters. 
//
// Copyright 2011-2016 by Yuri Ganshin, CMI, University of Wyoming, Laramie.
// http://www.uwyo.edu/cmi/
//
// This program is free sofware released using the Fast Light Tool Kit (FLTK)
// Copyright 1998-2010 by Bill Spitzak and others.
// For conditions of distribution and use, see the license agreement in 
//     http://www.fltk.org/COPYING.php
//
// K-means C++ routines Copyright 2009-2011 by John Burkardt
// 	https://people.sc.fsu.edu/~jburkardt/cpp_src/kmeans/kmeans.html
//
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
//
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Scroll.H>
#include <FL/fl_draw.H>	
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Native_File_Chooser.H>
#include <FL/Fl_File_Chooser.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Spinner.H>
#include <FL/Fl_Float_Input.H>
#include <FL/Fl_Input_Choice.H>
#include <FL/fl_show_colormap.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Box.H>
#include <FL/fl_message.H>
#include <FL/fl_ask.H> // for fl_alert() and fl_message()
#include <math.h>
//#include <fftw3.h>

using namespace std;

#define MARGIN		60
#define RIM			12
#define CURVE_BOX 80	// width of the box for curve's drawing
#define MIN_CURVE 2
#define MAX_CURVE	11
#define MAX_CLUST 10
//#define HDR_BAR	40
#define HYST_BINS	100	// Fixed number of bins in a hystogram
#define FLT_MAX	1E+37
#define FLT_MIN	-1E+37
#define MAX_ITER  40 // Maximum number of iterations for k-means algorithms
#define R8_HUGE  1.0E+30
//  The R8_HUGE value is NOT required to be the maximum representable R8,  
//  we simply want a "very large" but non-infinite number.
//
////////////////////////////////////////////////////////////////
/****               List of GLOBAL variables               ****/
char inpname[200];
int chosen=-1;
int Ncur=-1;						// Number of curves in file
int Ndat=-1;						// Number of samples in a curve
int canvas_w = 660, canvas_h = 960, win_bar = 8;  // pixels
float dep_sta,dep_end,dep_step,gain=1.25;
double *logs, *mini, *maxi;		// Min-Max for all curves
double *depth, *mean;
char curveID[MAX_CURVE][17], cgbuf[40];
int curve_no = 0;
Fl_Color bg_color = FL_WHITE;
//Fl_Color colr = 4;  // BLUE
Fl_Color colr[] = {40,60,62,72,80,120,222,136,144,192,200};
//int chooser = 0;
///////////////////////  CLUSTER RELATED
int *cluster;
int it_num, cluster_num=3, algor_no=1;
float *centroid, populat[MAX_CLUST], varian[MAX_CLUST];
char msg[MAX_CLUST][120];

////////////////////////////////////////////////////////////////
/****            List of Function Prototypes               ****/
//
void read_file_info(const char *name);
//
double *cluster_initialize ( int dim_num, int point_num, int cluster_num,
                            double point[], int *seed );
double *cluster_variance_compute (int dim_num,int point_num,int cluster_num,
                double point[], int cluster[], double cluster_center[] );
void kmeans_01 ( int dim_num, int point_num, int cluster_num, int it_max,
        int &it_num, double point[], int cluster[], double cluster_center[],
                int cluster_population[], double cluster_energy[] );
void kmeans_02 ( int dim_num, int point_num, int cluster_num, int it_max,
        int &it_num, double point[], int cluster[], double cluster_center[],
        int cluster_population[], double cluster_energy[] );
void i4vec_zero (int n, int a[]);;
void r8vec_zero ( int n, double a[] );
//
////////////////////////////////////////////////////////////////
/****         Global pointers to the GUI objects           ****/
////////////////////////////////////////////////////////////////
//
Fl_Double_Window *curve_window, *histo_window;
Fl_Window *opts_window;
Fl_Scroll *scroll;
Fl_Float_Input *min_inp, *max_inp;
Fl_Button *compute, *apply, *histo, *end_this;
Fl_Button *color1,*color2,*color3,*color4,*color5; 
Fl_Button *color6,*color7,*color8,*color9,*color10;
//
/////////////////////////////////////////////////////////////////////
/****          List of Widgets and Drawing functions            ****/
/////////////////////////////////////////////////////////////////////

/**********     File-choosing Menu    ************/
class File_menu : public Fl_Window {
    Fl_Native_File_Chooser *fc;
    
    void read_data(const char *filename)    // Process the chosen file
    {
        strcpy(inpname,filename);
        chosen=1;			// will wait while(chosen<0)
    }
    
    // Handle an 'Open' request from the menu
    static void open_cb(Fl_Widget *w, void *v) {
        File_menu *choice = (File_menu*)v;
        choice->fc->title("Open");
        choice->fc->type(Fl_Native_File_Chooser::BROWSE_FILE);
        switch ( choice->fc->show() ) {
            case -1: break;	// Error
            case  1: break; 	// Cancel
            default:		// choice->fc->show() returns 0
                choice->read_data(choice->fc->filename());
                // choice->deactivate();				
                break;
        }
    }
    
    static void quit_cb(Fl_Widget *w, void *v) {
        exit(0);
    }
    
public:
    File_menu() : Fl_Window(500,300,320,160,"Choose the file") {
        Fl_Menu_Bar *menu = new Fl_Menu_Bar(0,0,300,25);
        menu->add("&File/&Open",  FL_COMMAND+'o', open_cb, (void*)this);
        menu->add("&File/&Quit",  FL_COMMAND+'q', quit_cb);
        
        // Describe the chooser
        Fl_Box *box = new Fl_Box(20,45,w()-40,h()-65);
        box->color(FL_GRAY-1); 
        box->box(FL_FRAME_BOX);
        box->align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE|FL_ALIGN_WRAP);
        box->label("Choose TEXT-format file\nwith well-log data");
        box->labelsize(18);
        // Initialize the file chooser
        fc = new Fl_Native_File_Chooser();
        fc->filter("ASCII text files \t*.TXT, *.txt\n");			 
        end();
    }
};
/////////////////////////////////////////////////////////////////////
///////////    DRAW CURVES in a SCROLLABLE CANVAS       /////////////
/////////////////////////////////////////////////////////////////////
//  FLTK scrolling (with re-drawing) assumes (absolute) coordinates,
//  that must be calculated considering x() and y() values.
/////////////////////////////////////////////////////////////////////
class Draw_curves : public Fl_Widget 
{
    
public:
    Draw_curves(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}
    
    void draw() 
    {
        int i,j,dep;
        int x1, x2, y1, y2, xmin, xmax;
        double dx, val;
        Fl_Color bar_color = 15;
        
        fl_color(FL_WHITE);		// DRAW BG
        fl_rectf(x(),y(),w(),h());
        
        //cout << "X = " << x() <<	"; Y = " << y() << endl;
        //cout << "W = " << w() <<	"; H = " << h() << endl; // W = 460; H = 5561
        // x() and y() changes when scrolling, -> must be considered in coordinates
        // x()=y() = 0 in the top-left position, and become negative when scrolling
        
        fl_color(FL_BLACK);	fl_line_style(0,0);	fl_font(1,14);
        fl_draw("Depth",x()+8,y()+MARGIN/2);
        
        for(j=0;j<=Ncur;j++)  /////////////   Curve ID's
        {
            x1 = x() + j*CURVE_BOX + MARGIN;
            y1 = y();					// Zero coordinate
            y2 = y() + h();
            fl_yxline(x1, y1, y2);	// vertical line at x1 from y1 to y2
            
            if(j<Ncur) sprintf(cgbuf,"%s",curveID[j]);
            else sprintf(cgbuf,"%s","Cluster");
            
            y1 = y() + MARGIN/2;  
            fl_font(1,14);  fl_draw(cgbuf,x1+15,y1);  // draw text string
            
            if(j<Ncur)
            {
                y1 = y() + MARGIN - 10;  
                fl_font(1,12);  fl_draw("Min ---- Max",x1+6,y1);
            }
        }				
        
        fl_font(1,12);
        for(i=0;i<Ndat;i++)  ///////////   Depth Marks
        {
            dep = (int)depth[i];
            if(dep%10 ==0)			// 10-feet ticks
            {
                y1 = y() + i + MARGIN;
                x1 = x() + MARGIN;		x2 = x1 - 6;
                fl_xyline(x1, y1, x2);	// horizontal line at y1 from x1 to x2
                
                if(dep%100 ==0)	// 100-feet lines
                {
                    sprintf(cgbuf,"%5d",dep);
                    fl_draw(cgbuf, x()+12, y1+4);		// draw text string
                    x2 = x() + w();	fl_xyline(x1, y1, x2);
                }							
            }			
        }
        
        ///////////////////////////////////   Plot Log Curves	
        for(j=0;j<Ncur;j++)
        {
            x1 = x() + MARGIN + j*CURVE_BOX + 1;
            for(i=0;i<Ndat;i++)		// data samples
            {
                y1 = y() + MARGIN + i;
                val = logs[Ncur*i+j];
                dx = (val-mini[j])/(maxi[j]-mini[j]);				
                x2 = (int)(dx*CURVE_BOX) + x1;
                
                switch ( cluster[i] ) 
                {
                    case 0: bar_color = colr[1]; break;	
                    case 1: bar_color = colr[2]; break;
                    case 2: bar_color = colr[3]; break;
                    case 3: bar_color = colr[4]; break;
                    case 4: bar_color = colr[5]; break;
                    case 5: bar_color = colr[6]; break;
                    case 6: bar_color = colr[7]; break;
                    case 7: bar_color = colr[8]; break;
                    case 8: bar_color = colr[9]; break;
                    case 9: bar_color = colr[10]; break;	
                    default:	bar_color = 8; break;
                }
                
                fl_color(bar_color);
                fl_xyline(x1, y1, x2);  // Horizontal line from xmin to x1
            }
        }
        
        ///////////////////////////////////   Plot Cluster Bars		
        x1 = x() + MARGIN + Ncur*CURVE_BOX + 20;
        for(i=0;i<Ndat;i++)		// data samples
        {
            y1 = y() + MARGIN + i;			
            x2 = x1 + 40;			
            switch ( cluster[i] ) 
            {
                case  0: bar_color = colr[1]; break;
                case  1: bar_color = colr[2]; break;
                case  2: bar_color = colr[3]; break;
                case  3: bar_color = colr[4]; break;
                case  4: bar_color = colr[5]; break;
                case  5: bar_color = colr[6]; break;
                case  6: bar_color = colr[7]; break;
                case  7: bar_color = colr[8]; break;
                case  8: bar_color = colr[9]; break;
                case  9: bar_color = colr[10]; break;
                default:	bar_color = 8; break;
            }			
            fl_color(bar_color); fl_xyline(x1, y1, x2);
        }
        ///////////////////////////////////   Horizontal line
        x1 = x() + MARGIN;	x2 = x() + w();	y1 = y() + MARGIN;
        fl_color(FL_BLACK);	fl_xyline(x1, y1, x2);
        
    }	// void draw()
    
} *curve_plot;
//
/////////////////////////////////////////////////////////////////////
//////					HISTOGRAM  DRAWING					/////////
//////				in a double-buffered window				/////////
/////////////////////////////////////////////////////////////////////
class Draw_hist : public Fl_Double_Window 
{
	
public:
  	Draw_hist(int X,int Y,int W,int H) : Fl_Double_Window(X,Y,W,H) {}
	
	void draw(void)
	{
		Fl_Double_Window::draw();
		
		int i,j,count;
		int xx, xx1, ww, yy, yy1, hh, x_dim, y_dim;
		int bin_num=HYST_BINS;
		float bin_wid,hist[HYST_BINS];
		float val,min_val,max_val,Ymin,Ymax,max,scale;
		double pdf, q, sum, valx;
		double coef = sqrt(2.0*M_PI);
		
		x_dim = w() - (MARGIN+RIM);  y_dim = h() - (MARGIN+RIM);	
		/////////////////////////////////////////////////////////////
		fl_color(51);  fl_rectf(0,0,w(),h());			// Gray area
		fl_color(FL_BLACK); fl_rect(0,0,w(),h());		// bounding rect
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(MARGIN,RIM,x_dim,y_dim,255,255,255);
		
		/////////////////////////////////////////////////////////////
	
		min_val = (float)(mini[curve_no]*mean[curve_no]);  
		max_val = (float)(maxi[curve_no]*mean[curve_no]);								
		val = (float)((max_val-min_val));  // Round-up  Min-Max vales		
		if(val>1000.0)				
		{ 
			i=(int)(min_val/1000.0); min_val=1000.0*i; 
			i=(int)(max_val/1000.0); max_val=1000.0*(i+1);	
		}
		else
		if(val>100.0) 
		{ 
			i=(int)(min_val/100.0); min_val=100.0*i; 
			i=(int)(max_val/100.0); max_val=100.0*(i+1);
		}
		else
		if(val>10.0) 
		{ 
			i=(int)(min_val/10.0); min_val=10.0*i; 
			i=(int)(max_val/10.0); max_val=10.0*(i+1);
		}
		else
		if(val>1.0) 
		{ 
			i=(int)min_val; min_val=i; 
			i=(int)max_val; max_val=(i+1);
		}
		else						
		if(val>0.1) 
		{ 
			
			i=(int)(min_val*10.0); min_val=0.1*i; 
			i=(int)(max_val*10.0); max_val=0.1*(i+1);
		}
		else 
		{
			i=(int)(min_val*100.0); min_val=0.01*i; 
			i=(int)(max_val*100.0); max_val=0.01*(i+1);		
		}		
		//cout << "Min = " << min_val << ", Max = " << max_val << endl;				
		bin_wid = (max_val-min_val)/bin_num;

		
		for(i=0;i<bin_num;i++)				// 1.  Calculate Histogram
		{
			Ymin = min_val + bin_wid*i;
			Ymax = Ymin + bin_wid;
			count = 0;
			for(j=0;j<Ndat;j++)
			{
				val = (float)(logs[Ncur*j+curve_no]*mean[curve_no]);
				if(val<Ymin) continue;
				if(val>=Ymax) continue;
				count++;		// count == number of measurements in bin 'i'
			}
			hist[i] = (float)count;
		}
		
		max = 0.0;
		for(i=0;i<bin_num;i++)				// 2.  Normalize Histogram
		{
			val = hist[i]/(float)Ndat;		// Relative Frequency or Probability
			hist[i] = val;			
			if(val>max) max=val;
		}
		if(max<=0.0)
		{ fl_message("WARNING: No data found!\n\
						 Close the window to continue...");   return; }
		
		//  Round up Maximum Probability value (scale)
		if(max<.001) scale=.001;		
		else if(max<.005) scale=.005;
		else if(max<.01) scale=.01;
		else if(max<.05) scale=.05;
		else if(max<.1) scale=.1;
		else if(max<.5) scale=.5;
		else scale=1.0;
		
		////////////////////////////////////////////////////////
		yy1 = h() - MARGIN;	
		fl_line_style(0,1);	fl_font(1,12);  fl_color(FL_BLACK);
		///////////////////////////////////////////////////////
		//  X-axis ticks and labels
		for(i=0;i<10;i++)
		{
			val = min_val + (max_val-min_val)*i/10.0;
			xx = MARGIN + i*x_dim/10;
			fl_color(51);	fl_yxline(xx, RIM, yy1);  // light gray
			fl_color(0);	fl_yxline(xx, yy1, yy1+6);
			if(i%2==0)
			{ sprintf(cgbuf,"%g",val); fl_draw(cgbuf,xx-8,yy1+20); }
		}				
		
		fl_color(FL_BLUE);
		for(i=0;i<bin_num;i++)				// 2.  Plot Histogram
		{
			xx = MARGIN + i*x_dim/bin_num;
			xx1 = MARGIN + (i+1)*x_dim/bin_num;
			ww = xx1 - xx;
			hh = hist[i]*y_dim/scale; // Scaled by 'max'
			yy = yy1 - hh;	// 'y' must be smaller than 'h' in 'rectf'
			fl_rectf(xx, yy, ww, hh);
		}
		
		//////////////////////////////////////////////////////
		//  Y-axis ticks and labels
		fl_color(FL_BLACK);
		for(j=0;j<6;j++)
		{
			val = (100.0 - 20.0*j)*scale;	// Probability in percent
			yy = RIM + j*y_dim/5;
			fl_xyline(MARGIN-6, yy, MARGIN);	// horiz line
			sprintf(cgbuf,"%5.2f",val);	fl_draw(cgbuf,25,yy+4);
		}		
		fl_font(1,14);
		sprintf(cgbuf,"%s",curveID[curve_no]);
		fl_draw(cgbuf,x_dim*9/10,yy1+36);		
		fl_draw(90,"Probability [ % ]",18,y_dim/3);
		fl_font(0,12);
		sprintf(cgbuf,"Depth Range =  %.0f - %.0f",dep_sta,dep_end);
		fl_draw(cgbuf,MARGIN+x_dim/2,RIM+16);		
		sprintf(cgbuf,"Total number of measurements = %d",Ndat);
		fl_draw(cgbuf,MARGIN+x_dim/2,RIM+32);
		sprintf(cgbuf,"Number of bins =  %d",bin_num);
		fl_draw(cgbuf,MARGIN+40,RIM+16);		
		sprintf(cgbuf,"Bin width = %g",bin_wid);
		fl_draw(cgbuf,MARGIN+40,RIM+32);		
		
		fl_rect(MARGIN,RIM,x_dim,y_dim);
		
	} 	// void draw(void)
	
} *histo_plot;
//
////////////////////////////////////////////////////////////////
/************    List of Call Back functions    ***************/
////////////////////////////////////////////////////////////////

void input_curve(Fl_Widget*,void*data) {
    Fl_Input_Choice *in_curve=(Fl_Input_Choice *)data;
      
	for(curve_no=0; curve_no<Ncur; curve_no++) 
		if(strcmp(in_curve->value(),curveID[curve_no]) == 0) break;
	//cout << "Curve# = " << curve_no << endl;	 
}
//
void cluster_cb(Fl_Widget *w, void *) {
    Fl_Spinner *s = (Fl_Spinner *)w;	
    cluster_num = (int)s->value();
}
//
void algor_cb(Fl_Widget *w, void *) {
    Fl_Spinner *s = (Fl_Spinner *)w;	
    algor_no = (int)s->value();
}
//
/////////  Perform Cluster Analysis
//
void doCLUST(Fl_Widget *, void*)
{
    double *cluster_center;
    double *cluster_energy;
    double *cluster_variance;
    int *cluster_population;
    double cpp, cv;    
    int i, j;
    int seed = 123456789;
   
    //  Initialize cluster arrays
    //
    cluster_center = cluster_initialize(Ncur,Ndat,cluster_num,logs,&seed);
	
    cluster_energy = new double[cluster_num];
    cluster_population = new int[cluster_num];	
    
    for(i=0; i<Ndat; i++) cluster[i] = -1;
    
    switch (algor_no)
    {
        case 1:
            kmeans_01(Ncur, Ndat, cluster_num, MAX_ITER, it_num, logs,
                      cluster,cluster_center,cluster_population,cluster_energy);
            break;
        case 2:
            kmeans_02(Ncur, Ndat, cluster_num, MAX_ITER, it_num, logs,
                      cluster,cluster_center,cluster_population,cluster_energy);
            break;
        default: break;
    }
    
    for(j=0; j<cluster_num; j++)
        for(i=0;i<Ncur;i++) 
            centroid[i+j*Ncur] = (float)(cluster_center[i+j*Ncur]*mean[i]);
    
    cluster_variance = cluster_variance_compute(Ncur, Ndat, cluster_num,
                                            logs, cluster, cluster_center);
     
    for(j=0;j<MAX_CLUST;j++) sprintf(msg[j],"");
    
    for (j=0; j<cluster_num; j++)
    {
        i = cluster_population[j];
        cpp = ( ( double )( 100*i ) / ( double )Ndat );
        cv = cluster_variance[j];
        sprintf(msg[j],"%8d     %9.2f          %9.3f\n",j+1,cpp,cv);
        populat[j] = (float)cpp;
        varian[j] = (float)cv;
    }
 
    { fl_message("  Cluster statistics:\n\
       #   Population[%%]      Variance\n\%s%s%s%s%s%s%s%s%s%s\n\
   Number of iterations taken is %d\n\n\
Close the window to continue...",
        msg[0],msg[1],msg[2],msg[3],msg[4],
        msg[5],msg[6],msg[7],msg[8],msg[9],it_num); }
    

 	delete [] cluster_center;   
	delete [] cluster_energy;
	delete [] cluster_population;
	delete [] cluster_variance;
    
    apply->activate();
}
//
///////    Color Chooser    ////////////
//
void do_color1(Fl_Widget*, void*)
{	colr[1] = fl_show_colormap(bg_color);
    color1->color(colr[1]);  color1->redraw();
}
//
void do_color2(Fl_Widget*, void*)
{	colr[2] = fl_show_colormap(bg_color);
    color2->color(colr[2]);  color2->redraw();    
}
//
void do_color3(Fl_Widget*, void*)
{	colr[3] = fl_show_colormap(bg_color);
    color3->color(colr[3]);  color3->redraw();    
}
//
void do_color4(Fl_Widget*, void*)
{	colr[4] = fl_show_colormap(bg_color);
    color4->color(colr[4]);  color4->redraw();    
}
//
void do_color5(Fl_Widget*, void*)
{	colr[5] = fl_show_colormap(bg_color); 
    color5->color(colr[5]);  color5->redraw();    
}
//
void do_color6(Fl_Widget*, void*)
{	colr[6] = fl_show_colormap(bg_color);
    color6->color(colr[6]);  color6->redraw();    
}
//
void do_color7(Fl_Widget*, void*)
{	colr[7] = fl_show_colormap(bg_color);
    color7->color(colr[7]);  color7->redraw();    
}
//
void do_color8(Fl_Widget*, void*)
{	colr[8] = fl_show_colormap(bg_color);
    color8->color(colr[8]);  color8->redraw();    
}
//
void do_color9(Fl_Widget*, void*)
{	colr[9] = fl_show_colormap(bg_color);
    color9->color(colr[9]);  color9->redraw();    
}
//
void do_color10(Fl_Widget*, void*)
{	colr[10] = fl_show_colormap(bg_color); 
    color10->color(colr[10]);  color10->redraw();    
}
//
/////////  ReDRAW log data using cluster colors
//
void reDRAW(Fl_Widget *, void*)
{
    //chooser++; if(chooser>5) chooser=0;
    curve_plot->redraw();	
}
//
//
///////////    EXIT    ////////////
void exit_b(Fl_Widget *, void *) {
    exit(0);
}
//
///////    Save Cluster Info    ////////////
/**	The File_Chooser TYPE argument can be one of the following:
 * SINGLE - allows the user to select a single, existing file
 * MULTI - allows the user to select one or more existing files
 * CREATE - allows the user to select a single, existing file
 or specify a new filename
 * DIRECTORY - allows the user to select a single, existing directory
 **/
void create_outfile(Fl_Widget*, void*) 
{
    FILE *fp;
    int i,j;
    // Create the file chooser, and show it
    Fl_File_Chooser *fc = new Fl_File_Chooser(".",  // directory (current)
                            "*",                    // filter
                        Fl_File_Chooser::CREATE,    // chooser type
            "Select an existing or create a new file"); // title
    fc->show();
    
    while(fc->shown()) { Fl::wait(); } // Wait until user picks something
    
    if ( fc->value() == NULL ) return; // User hit cancel
    //cout << "File selected for output:" << fc->value() << endl;
    
    const char *clustname = fc->value();
    
    if ((fp = fopen(clustname, "a")) == NULL)
    { fl_message("ERROR opening file for writing\n\
                 Close the window to continue..."); return; }

    fprintf(fp,"Cluster statistics:\n"); 
	 
	 fprintf(fp,"\nNumber of iterations taken is %d\n\n",it_num); 
    fprintf(fp,"Cluster#   Population[%%]         Variance\n");   
    for(j=0;j<cluster_num;j++)
        fprintf(fp,"%8d     %9.2f          %9.3f\n",j+1,populat[j],varian[j]);
    fprintf(fp,"\n");    
    
    fprintf(fp,"Cluster Centroids:\n");
    for(i=0;i<Ncur;i++) fprintf(fp," %9s",curveID[i]); 
    fprintf(fp,"\n");
    for(j=0;j<cluster_num;j++)
    {
        fprintf(fp,"%2d",j+1);
        for(i=0;i<Ncur;i++) fprintf(fp," %9.3f",centroid[i+j*Ncur]);
        fprintf(fp,"\n");
    }
    fprintf(fp,"\n");
    
    fprintf(fp,"Depth  Cluster#\n");
    for(i=0;i<Ndat;i++)
        fprintf(fp,"%7.1f  %2d\n",depth[i],cluster[i]+1);
    
    fclose(fp);
}
//
void doHIST(Fl_Widget *, void*)
{  
	// Define Window Size with Margins
	int hist_w = 572;
	int hist_h = 472;
	
	histo_window = new Fl_Double_Window(600,200,hist_w,hist_h,
													"Histogram Window");
	histo_plot = new Draw_hist(0,0,hist_w,hist_h);
	
	histo_window->end();
	//curve_window->resizable(seis_window);	
	histo_window->show();
}
//
/////////////////////////////////////////////////////////////////////
/****   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   ****/
/////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
    int i,cur;
    double t,val;
    char cbuf1[300],cbuf2[600];	
    ///////////////////////////////////////////////////////////
    
    Fl::scheme(NULL); //  "plastic" for Aqua and "gtk+"
    // NULL use the scheme defined in the FLTK_SCHEME env.var. 
    
    if(argc == 2) strcpy(inpname,argv[1]);		// Do not use terminal
    else	// if you want to create an Apple  *.app bundle application
    {
        File_menu *choice = new File_menu();
        choice->show();
        choice->end(); // no more widgets will be added to the window
        while(chosen<0) Fl::wait();
        choice->hide();
    }
    
    read_file_info(inpname);
    
    cluster = new int[Ndat];  for(i=0; i<Ndat; i++) cluster[i] = -1;	
    
    opts_window = new Fl_Window(1050,50,420,780,"Settings & Control");
    
    Fl_Box* label1 = new Fl_Box(FL_FRAME_BOX,10,4,400,82,"");
    label1->color(6);	
    
    Fl_Box* label2 = new Fl_Box(80,12,250,20,"  File Information  ");
    label2->labelcolor(FL_BLACK);
    label2->labelfont(FL_HELVETICA_BOLD);	label2->labelsize(16);
    
    sprintf(cbuf1,"File contains %d curves,  %d samples each\n\
Starting depth = %.1f     Ending depth = %.1f\n\
Depth Increment = %.2f",
            Ncur,Ndat,dep_sta,dep_end,dep_step);
    Fl_Box* label3 = new Fl_Box(25,22,360,60,cbuf1);
    label3->labelfont(FL_HELVETICA);	label3->labelsize(14);
	 
	//////////////////////////////////////////////////////////////////
	////////////////////          HISTOGRAM         //////////////////
	//////////////////////////////////////////////////////////////////
	
	Fl_Input_Choice in_curve(120,104,90,24,"Select curve: ");
	in_curve.callback(input_curve, (void*)&in_curve);
	
	for(i=0;i<Ncur;i++) in_curve.add(curveID[i]);
	in_curve.value(curve_no);  // default curve for plotting
	in_curve.labelsize(14); in_curve.labelfont(FL_HELVETICA_BOLD);
	in_curve.align(Fl_Align(FL_ALIGN_LEFT));
	
	histo = new Fl_Button(225,100,155,30,"Display histogram");
	histo->color(62);						histo->labelcolor(FL_BLACK);
	histo->labelfont(3);					histo->labelsize(16);	
	histo->box(FL_RSHADOW_BOX);		histo->callback(doHIST);
	
	
    //////////////////////////////////////////////////////////////////
    ///////////    Customization for Cluster Analysis      ///////////
    //////////////////////////////////////////////////////////////////
    
    Fl_Box* label4 = new Fl_Box(50,135,300,40,
                        "Customize parameters for cluster analysis:");
    label4->labelfont(FL_HELVETICA_BOLD); label4->labelsize(16);		
    
    
    Fl_Spinner* spin_num = new Fl_Spinner(310, 175, 40, 24,
                            "Number of clusters to group data:  ");
    spin_num->labelsize(14);	spin_num->labelfont(FL_HELVETICA_BOLD);
    spin_num->minimum(2);		spin_num->maximum(MAX_CLUST);	
    spin_num->step(1);			spin_num->value(cluster_num);	
    spin_num->align(Fl_Align(FL_ALIGN_LEFT));
    spin_num->callback(cluster_cb);
    
    Fl_Spinner* spin_no = new Fl_Spinner(310, 210, 40, 24,
                                    "Clustering algorithm to use:  ");
    spin_no->labelsize(14);	spin_no->labelfont(FL_HELVETICA_BOLD);
    spin_no->minimum(1);		spin_no->maximum(2);	
    spin_no->step(1);			spin_no->value(algor_no);	
    spin_no->align(Fl_Align(FL_ALIGN_LEFT));
    spin_no->callback(algor_cb);
    
    Fl_Box* label5 = new Fl_Box(FL_FRAME_BOX,10,240,400,105,"");
    label5->color(6);		
    Fl_Box* label6 = new Fl_Box(80,245,250,20,"  References  ");
    label6->labelcolor(FL_BLACK);
    label6->labelfont(FL_HELVETICA_BOLD);	label6->labelsize(12);
    
    sprintf(cbuf2,
"Algorithm 1:                                                    \n\
Wendy L. Martinez and Angel R. Martinez. Computational Statistics\n\
Handbook with MATLAB. pp. 373-376, Chapman & Hall/CRC, 2002\n\
Algorithm 2:                                                    \n\
David N. Sparks. Algorithm AS 58: Euclidean Cluster Analysis. Journal\n\
of the Royal Statistical Society, Vol. 22, No. 1, pp. 126-130, 1973");
    Fl_Box* label7 = new Fl_Box(25,250,360,100,cbuf2);
    label7->labelfont(FL_HELVETICA);	label7->labelsize(11);	
    
    //////////////////////////////////////////////////////////////////////
    compute = new Fl_Button(115,365,180,30,"Cluster Analysis");
    compute->color(62);						compute->labelcolor(FL_BLACK);
    compute->labelfont(3);					compute->labelsize(18);
    compute->box(FL_RSHADOW_BOX);			compute->callback(doCLUST);	
    
    //////////////////////////////////////////////////////////////////////
    ////////////////   Choose cluster colors      ////////////////////////
    //////////////////////////////////////////////////////////////////////
    Fl_Box* label8 = new Fl_Box(125,403,180,40,
                                "Assign Cluster Colors:");
    label8->labelfont(FL_HELVETICA_BOLD);  label8->labelsize(16);
    
    color1 = new Fl_Button(100, 440, 80, 24, "Cluster 1");
    color1->labelcolor(FL_WHITE);		color1->color(colr[1]);
    color1->labelfont(3);				color1->labelsize(14);
    color1->callback(do_color1);      //color1->box(FL_RSHADOW_BOX);
    
    color2 = new Fl_Button(100, 475, 80, 24, "Cluster 2");
    color2->labelcolor(FL_WHITE);		color2->color(colr[2]);
    color2->labelfont(3);				color2->labelsize(14);
    color2->callback(do_color2);      //color2->box(FL_RSHADOW_BOX);
    
    color3 = new Fl_Button(100, 510, 80, 24, "Cluster 3");
    color3->labelcolor(FL_WHITE);		color3->color(colr[3]);
    color3->labelfont(3);				color3->labelsize(14);
    color3->callback(do_color3);      //color3->box(FL_RSHADOW_BOX);
    
    color4 = new Fl_Button(100, 545, 80, 24, "Cluster 4");
    color4->labelcolor(FL_WHITE);		color4->color(colr[4]);
    color4->labelfont(3);				color4->labelsize(14);
    color4->callback(do_color4);      //color4->box(FL_RSHADOW_BOX);
    
    color5 = new Fl_Button(100, 580, 80, 24, "Cluster 5");
    color5->labelcolor(FL_WHITE);		color5->color(colr[5]);
    color5->labelfont(3);				color5->labelsize(14);
    color5->callback(do_color5);      //color5->box(FL_RSHADOW_BOX);
    
    color6 = new Fl_Button(250, 440, 80, 24, "Cluster 6");
    color6->labelcolor(FL_WHITE);		color6->color(colr[6]);
    color6->labelfont(3);				color6->labelsize(14);
    color6->callback(do_color6);      //color6->box(FL_RSHADOW_BOX);
    
    color7 = new Fl_Button(250, 475, 80, 24, "Cluster 7");
    color7->labelcolor(FL_WHITE);		color7->color(colr[7]);
    color7->labelfont(3);				color7->labelsize(14);
    color7->callback(do_color7);      //color7->box(FL_RSHADOW_BOX);
    
    color8 = new Fl_Button(250, 510, 80, 24, "Cluster 8");
    color8->labelcolor(FL_WHITE);		color8->color(colr[8]);
    color8->labelfont(3);				color8->labelsize(14);
    color8->callback(do_color8);      //color8->box(FL_RSHADOW_BOX);
    
    color9 = new Fl_Button(250, 545, 80, 24, "Cluster 9");
    color9->labelcolor(FL_WHITE);		color9->color(colr[9]);
    color9->labelfont(3);				color9->labelsize(14);
    color9->callback(do_color9);      //color9->box(FL_RSHADOW_BOX);
    
    color10 = new Fl_Button(250, 580, 80, 24, "Cluster 10");
    color10->labelcolor(FL_WHITE);		color10->color(colr[10]);
    color10->labelfont(3);				color10->labelsize(14);
    color10->callback(do_color10);    //color10->box(FL_RSHADOW_BOX);
    
    //////////////////////////////////////////////////////////////////////
    apply = new Fl_Button(110,625,210,30,"Update Data Dispaly");
    apply->color(62);						apply->labelcolor(FL_BLACK);
    apply->labelfont(3);					apply->labelsize(18);
    apply->box(FL_RSHADOW_BOX);		apply->callback(reDRAW);
    apply->deactivate();
    
    //////////////////////////////////////////////////////////////
    ////////////    Save the Picks      //////////////////////////
    
    Fl_Menu_Bar* menubar = new Fl_Menu_Bar(140,675,150,30);
    menubar->add(" Save clusters to a file /Open",0,create_outfile);
    menubar->color(62);             menubar->labelcolor(FL_BLACK); 
    menubar->labelfont(3);              menubar->labelsize(18);
    menubar->box(FL_RSHADOW_BOX);   // menubar->deactivate();         
    //////////////////////////////////////////////////////////////
    
    end_this = new Fl_Button(165,725, 80, 30, "Exit");
    end_this->color(208);				end_this->labelcolor(FL_BLACK);
    end_this->labelfont(3);				end_this->labelsize(18);	
    end_this->box(FL_RSHADOW_BOX);	end_this->callback(exit_b);
    //////////////////////////////////////////////////////////////
    ///////////////    Log Display Window      ///////////////////
    
    int win_w = MARGIN + 400;
    int win_h = Fl::h() - 50;
    
    curve_window = new Fl_Double_Window(win_w,win_h,"Log Data Display");
    
    scroll = new Fl_Scroll(0,0,win_w,win_h);													
    curve_plot = new Draw_curves(0,0,(Ncur+1)*CURVE_BOX+MARGIN,Ndat+MARGIN);
    scroll->end();
    
    curve_window->end();
    curve_window->resizable(curve_window);	
    curve_window->show();		
    /////////////////////////////////////////////////////////////////
    
    opts_window->end(); 
    opts_window->show();
    
    return(Fl::run());
    return 0;
}
////////////////////////////////////////////////////////////////////////
/****     END_MAIN   END_MAIN   ENF_MAIN   END_MAIN   END_MAIN     ****/
////////////////////////////////////////////////////////////////////////
/****                Called Function Definition                    ****/
////////////////////////////////////////////////////////////////////////
void read_file_info(const char *name) // Reads ALL log-data into logs[]  
{
    //
    //   The input file may contain "#" to indicate a comment line
    //
    FILE *in_fp;
    int i,j;
    char *ptr, bigbuf[1000];	
    double val;
    
    if ((in_fp = fopen(name, "r")) == NULL)
    {fl_message("ERROR opening input file!\nClose the window to exit");exit(1);}
    
    fseek(in_fp, 0L, SEEK_SET);		// Go to beginning of the file
    while(fgets(bigbuf,999,in_fp))	// stop on first non-comment line
    { if(bigbuf[0]=='#') continue; else break; }
    
    ptr = strtok (bigbuf," ");		// get the first word from the 'bigbuf'
    // ptr = strtok (NULL, " ");  -> get word-2 and get tokens from word-1
    j=0;  
    while (ptr != NULL)
    {
        sscanf(ptr,"%lf",&val);
        ptr = strtok (NULL, " ");
        j++;	// j = curve index
    }
    
    Ncur = j-1;		// Get log data only, without DEPTH (in first column)
    
    if(Ncur>=MAX_CURVE || Ncur<MIN_CURVE)
    { fl_message("Number of curves should be from %d to %d\n\
Close the window to exit",MIN_CURVE,MAX_CURVE-1); fclose(in_fp); exit(1); }
    
    fseek(in_fp, 0L, SEEK_SET);		// Go to beginning of the file	
    Ndat=0;  
    while(fgets(bigbuf,999,in_fp)) { if(bigbuf[0]=='#') continue; Ndat++; }
    
    if(Ndat>100000 || Ndat<50)
    { fl_message("Number of samples should be from %d to %d\n\
Close the window to exit",50,100000); fclose(in_fp); exit(1); }
    ///////////////////////////////////////////////////////
    
    centroid = new float[Ncur*MAX_CLUST];
    depth = new double[Ndat];
    mean = new double[Ncur];
    mini = new double[Ncur];			// for logs scaling
    maxi = new double[Ncur];			// for logs scaling
    logs = new (nothrow) double[Ncur*Ndat];
    // 'nothrow' argument for operator 'new' returns NULL pointer on failure	
    if(logs == 0)
    { 
        fl_message("ERROR: NO memory to accomodate log data!\n\
Close the window to exit...");	
        fclose(in_fp);  exit(1);
    }
    ///////////////////////////////////////////////////////////////
    
    fseek(in_fp, 0L, SEEK_SET);		// Go to beginning of the file
    fgets(bigbuf,999,in_fp);			// First string
    if(bigbuf[0]=='#')
    {
        ptr = strtok (bigbuf+1," ");
        ptr = strtok (NULL, " ");	// Omit first token (Depth) after #
        j=0;  
        while (ptr != NULL)
        {
            sscanf(ptr,"%s",curveID[j]);
            ptr = strtok (NULL, " ");	j++;	// j = curve index
        }
    }
    else { for(j=0;j<Ncur;j++) sprintf(curveID[j],"Curve %d",j+1); }	
    ///////////////////////////////////////////////////////////////	
    
    fseek(in_fp, 0L, SEEK_SET);		// Go to beginning of the file
    i=0;										// and
    while(fgets(bigbuf,999,in_fp))	// READ DATA into logs[..]
    {
        if(bigbuf[0]=='#') continue;			
        ptr = strtok (bigbuf," ");  j=0;  
        while (ptr != NULL)
        {
            sscanf(ptr,"%lf",&val);
            if(j==0) depth[i] = val;
            else logs[Ncur*i+j-1] = val;
            ptr = strtok (NULL, " ");	j++;	// j = curve index
        }	
        
        if(j != (Ncur+1))   // together with DEPTH column
        { fl_message("Corrupted data at line %d\n\
Close the window to exit",i+1); fclose(in_fp); exit(1); }			
        
        i++;
        if(i>Ndat) 
        { fl_message("Exceeded number of lines 'Ndat' during data input\n\
Close the window to exit"); fclose(in_fp); exit(1); }		
    }
    fclose(in_fp);
    ///////////////////////////////////////////////////////////////
    // Calculate log mean values and divide all measureents by the corresponding
    // mean value, thus making log data independent of units of measurements 	
    for(j=0;j<Ncur;j++)
    {
        mean[j]=0.0;
        for(i=0;i<Ndat;i++)	// calculate mean's
        {
            mean[j] += logs[Ncur*i+j];
        }
        mean[j] /= (double)Ndat;
        
        for(i=0;i<Ndat;i++)	// make dimensionless log measurements
            logs[Ncur*i+j] /= mean[j];
    }
    ///////////////////////////////////////////////////////////////
    dep_sta = (float)depth[0];  dep_end = (float)depth[Ndat-1];
    dep_step = (float)(depth[1] - depth[0]);
    ///////////////////////////////////////////////////////////////	
    
    for(j=0;j<Ncur;j++)				// Search for Min-Max values 
    {
        mini[j]=FLT_MAX;  maxi[j]=FLT_MIN;
        for(i=0;i<Ndat;i++)
        { 
            val=logs[Ncur*i+j];
            if(val>maxi[j]) maxi[j]=val; 
            if(val<mini[j]) mini[j]=val;
        }
    }
}
//
//****************************************************************************80
double *cluster_initialize ( int dim_num, int point_num, int cluster_num, 
                            double point[], int *seed )
//****************************************************************************80
//
//   CLUSTER_INITIALIZE_5 initializes the cluster centers to random values.
//    In this case, each cluster center is a random convex combination 
//    of the data points.
//
//    08 October 2011, John Burkardt
//
{
	double *cluster_center;
	double column_sum;
	double *factor;
	int i,j,k;
	//
	factor = new double[point_num*cluster_num];
	//  factor[..] stores a doubly dimensioned matrix of pseudorandom R8's
	//    This routine implements the recursion:
	//      seed = 16807 * seed mod ( 2^31 - 1 )
	//      unif = seed / ( 2^31 - 1 )
	//  Reference:   Paul Bratley, Bennett Fox, Linus Schrage,
	//    A Guide to Simulation,  Springer Verlag, pages 201-202, 1983.

	for ( j = 0; j < cluster_num; j++ )
	{
		for ( i = 0; i < point_num; i++ )
		{
			k = *seed/127773;
			*seed = 16807*( *seed - k * 127773 ) - k * 2836;
			if(*seed<0) *seed += 2147483647;
			factor[i+j*point_num] = (double)(*seed)*4.656612875E-10;
		}
	}
	
	//  Make each column of factors have unit sum.
	//
	for ( k = 0; k < cluster_num; k++ )
	{
		column_sum = 0.0;
		for(j=0; j<point_num; j++) column_sum += factor[j+k*point_num];
		for(j=0; j<point_num; j++) factor[j+k*point_num] /= column_sum;
	}

	//  Set centers = points * factors.
	//
	cluster_center = new double[dim_num*cluster_num];
	for ( i = 0; i < dim_num; i++ )
	{
		for ( j = 0; j < cluster_num; j++ )
		{
			cluster_center[i+j*dim_num] = 0.0;
			for ( k = 0; k < point_num; k++ )
				cluster_center[i+j*dim_num]+=point[i+k*dim_num]*factor[k+j*point_num];
		}
	}	
												 											 
	delete [] factor;
	return cluster_center;
}
//****************************************************************************80

double *cluster_variance_compute ( int dim_num, int point_num, int cluster_num, 
  double point[], int cluster[], double cluster_center[] )
//****************************************************************************80
//
//   CLUSTER_VARIANCE_COMPUTE computes the variance of the clusters.
//    The cluster variance (from the cluster center) is the average of the 
//    sum of the squares of the distances of each point in the cluster to the 
//    cluster center.
//
//    06 October 2009;   John Burkardt
//
{
  int *cluster_population;
  double *cluster_variance;
  int i;
  int j;
  int k;
  double point_variance;

	cluster_population = new int[cluster_num];
	for(i=0; i<cluster_num; i++) cluster_population[i] = 0;
  
	cluster_variance = new double[cluster_num];
	for(i=0; i<cluster_num; i++) cluster_variance[i] = 0.0;
  

  for ( j = 0; j < point_num; j++ )
  {
    k = cluster[j];

    point_variance = 0.0;
    for ( i = 0; i < dim_num; i++ )
    {
      point_variance = point_variance +
        pow ( point[i+j*dim_num] - cluster_center[i+k*dim_num], 2 );
    }
    cluster_variance[k] = cluster_variance[k] + point_variance;
    cluster_population[k] = cluster_population[k] + 1;
  }

  for ( k = 0; k < cluster_num; k++ )
  {
    if ( 0 < cluster_population[k] )
    {
      cluster_variance[k] = cluster_variance[k] / cluster_population[k];
    }
  }

  delete [] cluster_population;
  return cluster_variance;
}
//****************************************************************************80
//****************************************************************************80

void kmeans_01 ( int dim_num, int point_num, int cluster_num, int it_max, 
  int &it_num, double point[], int cluster[], double cluster_center[], 
  int cluster_population[], double cluster_energy[] )
//****************************************************************************80
//
//   KMEANS_01 applies the H-Means algorithm.
//
//    The data for the H-Means problem is a set of N points X[..] in
//    M-dimensions, and a desired number of clusters K.
//
//    The goal is to determine K points Z[..], called cluster centers, so that
//    if we associate each point X[..] with its nearest Z[..] value, we minimize
//    the standard deviation or cluster energy.  Writing CLUSTER(I) to
//    indicate the index of the nearest cluster center to point X(I), the
//    energy can be written as:
//
//      Energy = Sum ( 1 <= I <= N ) || X(I) - Z(CLUSTER(I)) ||^2
//
//    where
//
//      || X - Z ||^2 = Sum ( 1 <= J <= M ) ( X(J) - Z(J) )^2
//
//    08 October 2011,  John Burkardt
//
//  Reference:
//  Wendy Martinez, Angel Martinez,  Computational Statistics Handbook
//  with MATLAB,  pages 373-376,  Chapman and Hall / CRC, 2002.
//
{
  double *centroid;
  int i;
  int j;
  int k;
  int k2;
  int missed;
  double point_energy;
  double point_energy_min;
  int swap;
//
//  On input, legal entries in CLUSTER are preserved, but
//  otherwise, each point is assigned to its nearest cluster.
//
  for ( j = 0; j < point_num; j++ )
  {
    if ( cluster[j] < 0 || cluster_num <= cluster[j] )
    {
      point_energy_min = R8_HUGE;

      for ( k = 0; k < cluster_num; k++ )
      {
        point_energy = 0.0;
        for ( i = 0; i < dim_num; i++ )
        {
          point_energy = point_energy + 
          pow ( point[i+j*dim_num] - cluster_center[i+k*dim_num], 2 );
        }

        if ( point_energy < point_energy_min )
        {
          point_energy_min = point_energy;
          cluster[j] = k;
        }
      }
    }
  }
  it_num = 0;

  while ( it_num < it_max )
  {
    it_num = it_num + 1;

//  #1:
//  Assign each point to the cluster of its nearest center.
//
    swap = 0;

    for ( j = 0; j < point_num; j++ )
    {
      point_energy_min = R8_HUGE;
      k = cluster[j];

      for ( k2 = 0; k2 < cluster_num; k2++ )
      {
        point_energy = 0.0;
        for ( i = 0; i < dim_num; i++ )
        {
          point_energy = point_energy + 
          pow ( point[i+j*dim_num] - cluster_center[i+k2*dim_num], 2 );
        }

        if ( point_energy < point_energy_min )
        {
          point_energy_min = point_energy;
          cluster[j] = k2;
        }
      }

      if ( k != cluster[j] )
      {
        swap = swap + 1;
      }
    }

//  Terminate if no points were swapped.
//
    if ( 1 < it_num ) { if ( swap == 0 ) break; }

//  #2:
//  Determine the total energy of the new clustering with current centroids.
//
    r8vec_zero ( cluster_num, cluster_energy );

    for ( j = 0; j < point_num; j++ )
    {
      k = cluster[j];

      point_energy = 0.0;
      for ( i = 0; i < dim_num; i++ )
      {
        point_energy = point_energy + 
        pow ( point[i+j*dim_num] - cluster_center[i+k*dim_num], 2 );
      }
      cluster_energy[k] = cluster_energy[k] + point_energy;
    }

//  #3:
//  Determine the centroids of the clusters.
//	 
    centroid = new double[dim_num*cluster_num];
    for(i=0; i<(dim_num*cluster_num); i++) centroid[i] = 0.0;	 
	 
    i4vec_zero ( cluster_num, cluster_population );

    for ( j = 0; j < point_num; j++ )
    {
      k = cluster[j];
      cluster_population[k] = cluster_population[k] + 1;
      for ( i = 0; i < dim_num; i++ )
      {
        centroid[i+k*dim_num] = centroid[i+k*dim_num] + point[i+j*dim_num];
      }
    }

//  Now divide by the population to get the centroid.
//  But if a center has no population, pick a point at random.
//
    missed = 0;

    for ( k = 0; k < cluster_num; k++ )
    {
      if ( cluster_population[k] != 0 )
      {
        for ( i = 0; i < dim_num; i++ )
        {
          centroid[i+k*dim_num] = centroid[i+k*dim_num]
          / ( double ) ( cluster_population[k] );
        }
      }
      else
      {
        for ( i = 0; i < dim_num; i++ )
        {
          centroid[i+k*dim_num] = point[i+missed*dim_num];
        }
        missed = missed + 1;
      }
    }

    for ( k = 0; k < cluster_num; k++ )
    {
      for ( i = 0; i < dim_num; i++ )
      {
        cluster_center[i+k*dim_num] = centroid[i+k*dim_num];
      }
    }

    delete [] centroid;

//  #4:
//  Determine the total energy of the current clustering with new centroids.
//
    r8vec_zero ( cluster_num, cluster_energy );

    for ( j = 0; j < point_num; j++ )
    {
      k = cluster[j];

      point_energy = 0.0;
      for ( i = 0; i < dim_num; i++ )
      {
        point_energy = point_energy + 
        pow ( point[i+j*dim_num] - cluster_center[i+k*dim_num], 2 );
      }
      cluster_energy[k] = cluster_energy[k] + point_energy;
    }
  }
  return;
}
//****************************************************************************80

void kmeans_02 ( int dim_num, int point_num, int cluster_num, int it_max, 
        int &it_num, double point[], int cluster[], double cluster_center[], 
            int cluster_population[], double cluster_energy[] )
//****************************************************************************80
//
//   KMEANS_02 applies the K-Means algorithm.
//
//    Given a matrix of POINT_NUM observations on DIM_NUM variables, the
//    observations are to be allocated to CLUSTER_NUM clusters in such 
//    a way that the within-cluster sum of squares is minimized.
//
//    09 October 2011
//    Original FORTRAN77 version by David Sparks.
//    C++ version by John Burkardt.
//
//  Reference:  David Sparks,  Algorithm AS 58: 
//    Euclidean Cluster Analysis,  Applied Statistics,
//    Volume 22, Number 1, 1973, pages 126-130.
//
{
	double dc;
	double de;
	double *f;
	int i;
	int il;
	int ir;
	int j;
	int j2;
	int k;
	double point_energy;
	double point_energy_min;
	int swap;
	
	it_num = 0;

	//  For each observation, calculate the distance from each cluster
	//  center, and assign to the nearest.
	//
	for ( j = 0; j < point_num; j++ )
	{
		point_energy_min = R8_HUGE;
		cluster[j] = -1;
		
		for ( k = 0; k < cluster_num; k++ )
		{
			point_energy = 0.0;
			for ( i = 0; i < dim_num; i++ )
			{
				point_energy = point_energy + 
				pow ( point[i+j*dim_num] - cluster_center[i+k*dim_num], 2 );
			}
			
			if ( point_energy < point_energy_min )
			{
				point_energy_min = point_energy;
				cluster[j] = k;
			}
		}
	}
	
	//  Determine the cluster population counts.
	//
	i4vec_zero ( cluster_num, cluster_population );
	
	for ( j = 0; j < point_num; j++ )
	{
		k = cluster[j];
		cluster_population[k] = cluster_population[k] + 1;
	}
	
	//  Calculate the mean and sum of squares for each cluster.
	//
	r8vec_zero ( dim_num * cluster_num, cluster_center );
	
	for ( j = 0; j < point_num; j++ )
	{
		k = cluster[j];
		for ( i = 0; i < dim_num; i++ )
		{
			cluster_center[i+k*dim_num] = cluster_center[i+k*dim_num] 
			+ point[i+j*dim_num];
		}
	}
	
	for ( k = 0; k < cluster_num; k++ )
	{
		if ( 0 < cluster_population[k] )
		{
			for ( i = 0; i < dim_num; i++ )
			{
				cluster_center[i+k*dim_num] = cluster_center[i+k*dim_num] / 
				( double ) cluster_population[k];
			}
		}
	}
	
	//  Set the point energies.
	//
	f = new double[point_num];
	
	for ( j = 0; j < point_num; j++ )
	{
		k = cluster[j]; f[j] = 0.0;
		for ( i = 0; i < dim_num; i++ )
		{
			f[j] += pow( point[i+j*dim_num]-cluster_center[i+k*dim_num], 2 );
		}
	}
	
	//  Set the cluster energies.
	//
	r8vec_zero ( cluster_num, cluster_energy );
	
	for ( j = 0; j < point_num; j++ )
	{
		k = cluster[j];
		cluster_energy[k] = cluster_energy[k] + f[j];
	}
	
	//  Adjust the point energies by a weight factor.
	//
	for ( j = 0; j < point_num; j++ )
	{
		k = cluster[j];
		if ( 1 < cluster_population[k] )
		{
			f[j] = f[j] * ( double ) ( cluster_population[k] ) 
			/ ( double ) ( cluster_population[k] - 1 );
		}
	}
	
	//  Examine each observation in turn to see if it should be
	//  reassigned to a different cluster.
	//
	it_num = 0;
	
	while ( it_num < it_max )
	{
		it_num = it_num + 1;
		
		swap = 0;
		
		for ( j = 0; j < point_num; j++ )
		{
			il = cluster[j];
			ir = il;
			
			if ( cluster_population[il] <= 1 ) continue;
			
			dc = f[j];
			
			for ( k = 0; k < cluster_num; k++ )
			{
				if ( k != il )
				{
					de = 0.0;
					for ( i = 0; i < dim_num; i++ )
					{
                        de=de+pow(point[i+j*dim_num]-cluster_center[i+k*dim_num],2);
					}
					de = de * ( double )cluster_population[k] 
					/ ( double )( cluster_population[k] + 1 );
					
					if ( de < dc ) { dc = de;  ir = k; }
				}
			}
			
			//  If the lowest value was obtained by staying in the current cluster,
			//  then cycle.
			//
			if( ir == il ) continue;
			
			//  Reassign the point from cluster IL to cluster IR.
			//
			for ( i = 0; i < dim_num; i++ )
			{
				cluster_center[i+il*dim_num] = ( cluster_center[i+il*dim_num]
                * ( double ) ( cluster_population[il] ) - point[i+j*dim_num] )
				/ ( double ) ( cluster_population[il] - 1 );
				
				cluster_center[i+ir*dim_num] = ( cluster_center[i+ir*dim_num]
                * ( double ) ( cluster_population[ir] ) + point[i+j*dim_num] )
				/ ( double ) ( cluster_population[ir] + 1 );
			}
			cluster_energy[il] = cluster_energy[il] - f[j];
			cluster_energy[ir] = cluster_energy[ir] + dc;
			cluster_population[ir] = cluster_population[ir] + 1;
			cluster_population[il] = cluster_population[il] - 1;
			
			cluster[j] = ir;
			
			//  Adjust the value of F for points in clusters IL and IR.
			//  
			for ( j2 = 0; j2 < point_num; j2++ )
			{
				k = cluster[j2];
				
				if ( k == il || k == ir )
				{
					f[j2] = 0.0;
					for ( i = 0; i < dim_num; i++ )
					{
						f[j2] = f[j2]+pow(point[i+j2*dim_num]-cluster_center[i+k*dim_num],2);
					}
					
					if ( 1 < cluster_population[k] )
					{
						f[j2] = f[j2] * ( double ) ( cluster_population[k] ) 
						/ ( ( double ) ( cluster_population[k] - 1 ) );
					}
				}
			}
			swap = swap + 1;
		}
		//  Exit if no reassignments were made during this iteration.
		//
        if( swap == 0 ) break;
	}	
	//  Compute the cluster energies.
	//
	r8vec_zero ( cluster_num, cluster_energy );
	
	for ( j = 0; j < point_num; j++ )
	{
		k = cluster[j];
		
		point_energy = 0.0;
		for ( i = 0; i < dim_num; i++ )
		{
			point_energy = point_energy + 
			pow ( point[i+j*dim_num] - cluster_center[i+k*dim_num], 2 );
		}
		cluster_energy[k] = cluster_energy[k] + point_energy;
	}	
	delete [] f;	
	return;
}
//
void i4vec_zero(int n, int a[])  // zeroes a vector of I4's.
{
  int i;
  for (i=0; i<n; i++) a[i] = 0;
  return;
}
//
void r8vec_zero (int n, double a[])  // zeroes a vector of R8's
{
	int i;
	for(i=0; i<n; i++)  a[i] = 0.0;
	return;
}
